/*
 * Modern, New Agey LED Clock made with SparkFun's Pro Micro Board
 *
 * HARDWARE:
 *  For display: an LPD8806 LED strip containing 14 RGB LEDs. One LED for each hour
 *               plus two LEDs for misc usage. Available from adafruit.com.
 *               pin 4 is data to LEDs; pin 5 is clock to LEDs
 *
 *  For time keeping: a ChronoDot module based on the DS3231 RTC chip. Accurate to
 *                    one minute a year. Available from adafruit.com.
 *                    pin 2 is data to RTC; pin 3 is clock to RTC
 *
 *  For UI: two minature SPST momentary contact pushbutton switches.
 *  pin 7 is inc button; pin 6 is dec button
 *
 * Hardware and software design by: Craig A. Lindley
 * Code compiles on Arduino 1.0 or newer only
 *
 * Revision History
 *
 * Version: 1.0 - 05/06/2012
 *   Initial release
 * Version: 2.0 - 05/12/2012
 *   Changed Chronodot to run in 12 hour format
 *   Eliminated use of Chronodot library
 *   Changed hour setting to AM/PM format 
 */

#include <Arduino.h> 
#include <Bounce.h>
#include <Wire.h>

#include "datatypes.h"

#define DEBUG false

// Button Definitions
// Input pin assignments
#define INC_BUTTON_PIN   7
#define DEC_BUTTON_PIN   6

// Instantiate Bounce objects for debouncing push buttons
Bounce incButton = Bounce(INC_BUTTON_PIN, 5); 
Bounce decButton = Bounce(DEC_BUTTON_PIN, 5); 

// LED Strip Definitions
// Count of LEDs
#define NUMBER_OF_LEDS  14

// Output pin assignments
#define LED_CLOCK_PIN    5
#define LED_DATA_PIN     4

// Masks and pointers to speed access
uint8_t clockPinMask, dataPinMask;
volatile uint8_t *clockPort, *dataPort;

// Array of LED data in GRB order. Component data is 7 bits wide.
uint32_t LEDData[NUMBER_OF_LEDS];

// Brightness correction data for LEDs
// See SCurve.java for how this data was generated
uint8_t GAMMA [] = {
  0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,
  1,2,2,2,2,2,2,3,3,3,4,4,4,5,5,6,
  6,7,7,8,9,9,10,11,12,13,14,16,17,18,20,21,
  23,25,27,29,31,33,36,38,41,43,46,49,52,55,58,61,
  64,66,69,72,75,78,81,84,86,89,91,94,96,98,100,102,
  104,106,107,109,110,111,113,114,115,116,117,118,118,119,120,120,
  121,121,122,122,123,123,123,124,124,124,125,125,125,125,125,125,
  126,126,126,126,126,126,126,126,126,126,126,126,127,127,127,127
};

// Create the LED colors for the various settings
uint32_t hourColor   = createPackedColor(127,  0,  0);  // Red
uint32_t minuteColor = createPackedColor(  0,127,  0);  // Green
uint32_t monthColor  = createPackedColor(127,127,  0);  // Yellow
uint32_t dateColor   = createPackedColor(127,  0,127);  // Magenta
uint32_t yearColor   = createPackedColor(  0,127,127);  // Cyan

// Misc colors
uint32_t BLACK = createPackedColor(0,0,0);
uint32_t BLUE  = createPackedColor(0,0,63);
uint32_t RED   = createPackedColor(63,0,0);

// Real Time Clock Definitions
#define CHRONODOT_I2C_ADDRESS    104
#define CHRONODOT_SECONDS_REGISTER 0
#define CHRONODOT_MINUTES_REGISTER 1
#define CHRONODOT_HOURS_REGISTER   2
#define CHRONODOT_DAY_REGISTER     3
#define CHRONODOT_DATE_REGISTER    4
#define CHRONODOT_MONTH_REGISTER   5
#define CHRONODOT_YEAR_REGISTER    6
#define CHRONODOT_STATUS_REGISTER 15

uint8_t seconds;  // 0..59
uint8_t minutes;  // 0..59
uint8_t hours;    // 1..12
uint8_t date;     // 1..31
uint8_t month;    // 1..12
uint8_t year;     // 0..99
uint8_t pmAm;     // 0 == AM: 1 == PM

boolean secondsToggle;
boolean hourMinuteToggle;
uint8_t latchedMinute;

// Overall state of the clock
CLOCK_STATE clockState;

/************************************************************************/
/**************************** Button Functions **************************/
/************************************************************************/

// Returns the current state of both pushbuttons
BUTTON_STATE readButtonState() {

  BUTTON_STATE bState;

  // Update the switches states
  incButton.update();
  decButton.update();

  // Read the state of both buttons
  boolean incButtonState = (incButton.read() == LOW);
  boolean decButtonState = (decButton.read() == LOW);

  // Neither button was pressed so just return indication
  if (!incButtonState && !decButtonState) {
    return NEITHER;
  }  
  // Both buttons were pressed simultaneously
  if (incButtonState && decButtonState) {
    bState = BOTH;
  }  
  else  {
    // If we get here one of the buttons was pressed
    // Delay to see if user pressed both buttons at slightly
    // different times.
    delay(50);

    // Update the switches states again
    incButton.update();
    decButton.update();

    // Read the state of both buttons
    incButtonState = (incButton.read() == LOW);
    decButtonState = (decButton.read() == LOW);

    // Now look for the final state
    if (incButtonState && decButtonState) {
      bState = BOTH;
    }  
    else if (incButtonState && !decButtonState) {
      bState = PLUS;
    }  
    else  {
      bState =  MINUS;
    }  
  }

  // Now wait until user releases all buttons
  do {
    // Update the switches states
    incButton.update();
    decButton.update();

    // Read the state of both buttons
    // Button press == HIGH
    incButtonState = (incButton.read() == LOW);
    decButtonState = (decButton.read() == LOW);
  } 
  while (incButtonState || decButtonState);

  if (DEBUG) {
    Serial.print("Button state: ");
    Serial.println(bState);
  }

  // Return detected state
  return bState;
}

/************************************************************************/
/*************************** LED strip functions ************************/
/************************************************************************/

// Convert separate R,G,B into packed 32-bit GRB color
uint32_t createPackedColor(uint8_t r, uint8_t g, uint8_t b) {
  return 0x808080 | ((uint32_t) g << 16) | ((uint32_t) r << 8) | (uint32_t) b;
}

// Set LED color from separate 7-bit R, G, B components
void setLEDColor(uint16_t ledNumber, uint8_t r, uint8_t g, uint8_t b) {
  uint32_t data = (g | 0x80);
  data <<= 8;
  data |= (r | 0x80);
  data <<= 8;
  data |= (b | 0x80);
  LEDData[ledNumber] = data;
}

// Set LED color from RGB struct
void setLEDColor(uint16_t ledNumber, struct RGB color) {
  setLEDColor(ledNumber, color.red, color.green, color.blue);  
}

// Set LED color from packed RGB value
void setLEDColor(uint16_t ledNumber, uint32_t c) {
  uint32_t data = ((c>>16) | 0x80);
  data <<= 8;
  data |= ((c>>8) | 0x80);
  data <<= 8;
  data |= ((c) | 0x80);
  LEDData[ledNumber] = data;
}

// Set all LEDs to black
void clearLEDs() {
  for (int i = 0; i < NUMBER_OF_LEDS; i++) {
    setLEDColor(i, BLACK);
  }
  show();
}

// Write a byte of LED data to LED strip
void writeByte(byte b) {

  for (uint8_t bit = 0x80; bit; bit >>= 1) {      
    if (b & bit) {
      *dataPort |=  dataPinMask;
    }  
    else {
      *dataPort &= ~dataPinMask;
    }
    *clockPort |=  clockPinMask;
    *clockPort &= ~clockPinMask;
  }
}

// Push LED data to LED strip
void show() {

  // Get the strip's attention
  writeByte(0);

  // Write 24 bits/LED
  for (int i = 0; i < NUMBER_OF_LEDS; i++) {
    uint32_t pixel = LEDData[i];
    writeByte(pixel >> 16 & 0xFF);
    writeByte(pixel >> 8  & 0xFF);
    writeByte(pixel       & 0xFF);
  }
  // Finish up
  writeByte(0);
}

// Color space conversion
// Input arguments
// hue in degrees (0 - 360.0)
// saturation in percent (0.0 - 1.0)
// value in percent (0.0 - 1.0)
// Output arguments
// red, green blue (0.0 - 1.0)
void _HSVtoRGB(double hue, double saturation, double value, double *red, double *green, double *blue) {

  int i;
  double f, p, q, t;
  if (saturation == 0) {
    // achromatic (grey)
    *red = *green = *blue = value;
    return;
  }
  hue /= 60;			// sector 0 to 5
  i = floor(hue);
  f = hue - i;			// factorial part of h
  p = value * (1 - saturation);
  q = value * (1 - saturation * f);
  t = value * (1 - saturation * (1 - f));
  switch(i) {
  case 0:
    *red = value;
    *green = t;
    *blue = p;
    break;
  case 1:
    *red = q;
    *green = value;
    *blue = p;
    break;
  case 2:
    *red = p;
    *green = value;
    *blue = t;
    break;
  case 3:
    *red = p;
    *green = q;
    *blue = value;
    break;
  case 4:
    *red = t;
    *green = p;
    *blue = value;
    break;
  default:
    *red = value;
    *green = p;
    *blue = q;
    break;
  }
}

// Create fully saturated color. Store in RGB structure
void createColor(float hue, float value, struct RGB *color) {
  double r, g, b;

  _HSVtoRGB(hue, 1.0, value, &r, &g, &b);
  (*color).red   = (uint8_t) (r * 127);
  (*color).green = (uint8_t) (g * 127);
  (*color).blue =  (uint8_t) (b * 127);
}

// This is called to alert user to mode change
// Flashes the 12 LEDS around the clock in speified color.
void displayFlash(uint32_t flashColor) {

  for (int ledNum = 2; ledNum < NUMBER_OF_LEDS; ledNum++) {
    setLEDColor(ledNum, flashColor);
    show();
    delay(50);
    setLEDColor(ledNum, BLACK);
    show();
    delay(50);
  }
  delay(1000);
}

// Set the background color when in clock mode
void displayBackground() {

  for (int ledNum = 2; ledNum < NUMBER_OF_LEDS; ledNum++) {
    setLEDColor(ledNum, BLUE);
  }
  show();
}

// Rotating color palette pattern
void rotatorPattern() {

  RGB color;
  double hue;

  clearLEDs();

  // Hue increment
  double hueInc = 360.0 / 12.0;

  // Repeat the pattern specified times
  for (int i = 0; i < 20; i++) {
    // Pattern rotates 12 times so each color has been in each position
    for (int s = 0; s < 12; s++) {
      for (int led = 0; led < NUMBER_OF_LEDS - 2; led++) {
        hue = ((int)((led + s) * hueInc)) % 360;
        createColor(hue, 0.5, &color);
        setLEDColor(led + 2, color);
      }
      show();    
      delay(250);
    } 
  }
  clearLEDs(); 
}

// Brightening colors pattern
void brighteningPattern() {

  RGB color;
  double value;

  clearLEDs();

  // Degrees in HSV color space for primary colors
  float redHue   =   0;
  float greenHue = 120;
  float blueHue  = 240;

  int brightnessLevels = 128;
  int brightnessTimeMS = 200;

  int paletteEntries = 128;

  double hueIncrement = 360.0 / paletteEntries;

  // Red: Gradually increase value
  for (int s = 0; s < brightnessLevels; s++) {
    value = GAMMA[s] / 127.0;

    for (int led = 0; led < NUMBER_OF_LEDS - 2; led++) {
      createColor(redHue, value, &color);
      setLEDColor(led + 2, color);
    }
    show();    
    delay(brightnessTimeMS);
  }  

  double hue = redHue;  
  for (int p = 0; p < paletteEntries; p++) {    
    for (int led = 0; led < NUMBER_OF_LEDS - 2; led++) {
      createColor(hue, 1.0, &color);
      setLEDColor(led + 2, color);
    }
    show();
    hue += hueIncrement;
    if (hue >= 360.0) {
      hue -= 360.0;
    } 
    delay(brightnessTimeMS);    
  }

  clearLEDs(); 

  // Green: Gradually increase value
  for (int s = 0; s < brightnessLevels; s++) {
    value = GAMMA[s] / 127.0;

    for (int led = 0; led < NUMBER_OF_LEDS - 2; led++) {
      createColor(greenHue, value, &color);
      setLEDColor(led + 2, color);
    }
    show();    
    delay(brightnessTimeMS);
  }

  hue = greenHue;  
  for (int p = 0; p < paletteEntries; p++) {    
    for (int led = 0; led < NUMBER_OF_LEDS - 2; led++) {
      createColor(hue, 1.0, &color);
      setLEDColor(led + 2, color);
    }
    show();
    hue += hueIncrement;    
    if (hue >= 360.0) {
      hue -= 360.0;
    } 
    delay(brightnessTimeMS);    
  }

  clearLEDs(); 

  // Blue: Gradually increase value
  for (int s = 0; s < brightnessLevels; s++) {
    value = GAMMA[s] / 127.0;

    for (int led = 0; led < NUMBER_OF_LEDS - 2; led++) {
      createColor(blueHue, value, &color);
      setLEDColor(led + 2, color);
    }
    show();    
    delay(brightnessTimeMS);
  }

  hue = blueHue;  
  for (int p = 0; p < paletteEntries; p++) {    
    for (int led = 0; led < NUMBER_OF_LEDS - 2; led++) {
      createColor(hue, 1.0, &color);
      setLEDColor(led + 2, color);
    }
    show();
    hue += hueIncrement;    
    if (hue >= 360.0) {
      hue -= 360.0;
    } 
    delay(brightnessTimeMS);    
  }

  clearLEDs(); 
}

/************************************************************************/
/*********************** Chronodot Access Functions *********************/
/************************************************************************/

byte decToBcd(byte num) {
  return (((num / 10) << 4) | (num % 10));
}

byte bcdToDec(byte num) {
  return (((num >> 4) * 10) + (num & 0x0F));
}

// Determine if the Chronodot is running by examining the status register.
// If bit 7 is high the oscillator is stopped which is bad
// Returns true if RTC is running; false otherwise
boolean rtcRunning() {

  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write((byte) CHRONODOT_STATUS_REGISTER);
  Wire.endTransmission();
  Wire.requestFrom(CHRONODOT_I2C_ADDRESS, 1);
  uint8_t status = Wire.read();
  return (status & (1 << 7)) == 0;
}

// Set the RTC into 12 hour format
void set12HourFormat() {

  // Read the hours register
  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write((byte) CHRONODOT_HOURS_REGISTER);
  Wire.endTransmission();
  Wire.requestFrom(CHRONODOT_I2C_ADDRESS, 1);
  uint8_t temp = Wire.read();

  temp |= (1 << 6);  // Set bit to enable 12 hour format

  // Write the modified hours value back to the RTC   
  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write(CHRONODOT_HOURS_REGISTER);
  Wire.write(temp);
  Wire.endTransmission();
}

// Set the time to the RTC in 12 hour format
void setTime() {

  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write((byte) CHRONODOT_SECONDS_REGISTER);
  Wire.write(decToBcd(seconds));
  Wire.write(decToBcd(minutes));

  // Process hours value for 12 hour format
  uint8_t data = decToBcd(hours);

  // Is the time in the PM ?
  if (pmAm) {
    // Yes, so set the Pm/Am bit
    data |= (1 << 5);
  }  
  else  {
    // No, so reset the Pm/Am bit
    data &= ~(1 << 5);    
  }
  // Finally set the 12/24 format bit 
  data |= (1 << 6);

  // Write the modified hours value
  Wire.write(data);
  Wire.endTransmission();
}

// Get the time from the RTC in 12 hour format
void getTime() {

  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write((byte) CHRONODOT_SECONDS_REGISTER);
  Wire.endTransmission();
  Wire.requestFrom(CHRONODOT_I2C_ADDRESS, 3);
  seconds = bcdToDec(Wire.read() & 0x7F);
  minutes = bcdToDec(Wire.read() & 0x7F);

  // Read the hours value
  uint8_t temp = Wire.read();

  // Is it AM or PM
  pmAm = ((temp & (1 << 5)) != 0);

  // Mask off 12/24 and Pm/Am bits to get hours value
  hours = bcdToDec(temp & 0x1F);
} 

// Set the date to the RTC
void setDate() {

  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write(CHRONODOT_DATE_REGISTER);
  Wire.write(decToBcd(date));
  Wire.write(decToBcd(month));
  Wire.write(decToBcd(year));
  Wire.endTransmission();
}

// Get the date from the RTC
void getDate() {

  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write(CHRONODOT_DATE_REGISTER);
  Wire.endTransmission();
  Wire.requestFrom(CHRONODOT_I2C_ADDRESS, 3);
  date  = bcdToDec(Wire.read());
  month = bcdToDec(Wire.read());
  year  = bcdToDec(Wire.read());
}

// Print out the current time from the RTC
void printDateTime() {  

  if(hours < 10) {
    Serial.print("0");
  }
  Serial.print(hours, DEC);
  Serial.print(':');

  if(minutes < 10) {
    Serial.print("0");
  }
  Serial.print(minutes, DEC);
  Serial.print(':');

  if(seconds < 10) {
    Serial.print("0");
  }
  Serial.print(seconds, DEC);

  Serial.print(' ');

  if (pmAm) {
    Serial.print("PM");
  }  
  else  {
    Serial.print("AM");
  }

  Serial.print(' ');

  if(month < 10) {
    Serial.print("0");
  }
  Serial.print(month, DEC);
  Serial.print('/');

  if(date < 10) {
    Serial.print("0");
  }
  Serial.print(date, DEC);
  Serial.print('/');

  Serial.print(year + 2000, DEC);
  Serial.println();
}

/************************************************************************/
/************************** Display Functions ***************************/
/************************************************************************/

// Valid hours are 1 .. 12
uint8_t HOUR_LED_MAP [] = {
  8, 9, 10, 11, 12, 13, 2, 3, 4, 5, 6, 7, 8
};

// Display hour value plus Am/Pm indication
void displayHour() {

  // Set all LEDs in strip to black color
  clearLEDs();

  setLEDColor(HOUR_LED_MAP[hours], hourColor);

  // Time in the PM ?
  if (pmAm) {
    // PM, set led 1
    setLEDColor(1, RED);
  }  
  else  {
    // AM, set led 0
    setLEDColor(0, RED);
  }
  show();
}

void displayMinute() {

  // Set all LEDs in strip to black color
  clearLEDs();

  int fullCycles = minutes / 12;
  int partialCycle = minutes % 12;

  for (int i = 0; i < fullCycles; i++) {
    for (int led = 2; led < NUMBER_OF_LEDS; led++) {
      setLEDColor(led, minuteColor);
    }
    show();
    delay(250);
    clearLEDs();
    delay(400);
  }
  setLEDColor(HOUR_LED_MAP[partialCycle], minuteColor);  
  show();
}

void displayMonth() {

  // Set all LEDs in strip to black color
  clearLEDs();

  setLEDColor(HOUR_LED_MAP[month], monthColor);
  show();
}

void displayDate() {

  // Set all LEDs in strip to black color
  clearLEDs();

  int fullCycles = date / 12;
  int partialCycle = date % 12;

  for (int i = 0; i < fullCycles; i++) {
    for (int led = 2; led < NUMBER_OF_LEDS; led++) {
      setLEDColor(led, dateColor);
    }
    show();
    delay(250);
    clearLEDs();
    delay(400);
  }
  setLEDColor(HOUR_LED_MAP[partialCycle], dateColor);  
  show();
}

void displayYear() {

  // Set all LEDs in strip to black color
  clearLEDs();

  int fullCycles = year / 12;
  int partialCycle = year % 12;

  for (int i = 0; i < fullCycles; i++) {
    for (int led = 2; led < NUMBER_OF_LEDS; led++) {
      setLEDColor(led, yearColor);
    }
    show();
    delay(250);
    clearLEDs();
    delay(400);
  }
  setLEDColor(HOUR_LED_MAP[partialCycle], yearColor);  
  show();
}

// Display the complete date on the clock
void displayFullDate() {

  displayMonth();
  delay(3000);
  displayDate();
  delay(3000);
  displayYear();
  delay(3000);
}

/************************************************************************/
/********** Setup all subsystems before clock starts operation **********/
/************************************************************************/

void setup() {

  if (DEBUG) {
    // Setup serial link for debugging
    Serial.begin(115200);
  }

  // Setup push buttons as inputs with pullups
  pinMode(INC_BUTTON_PIN, INPUT);
  digitalWrite(INC_BUTTON_PIN, HIGH);
  pinMode(DEC_BUTTON_PIN, INPUT);
  digitalWrite(DEC_BUTTON_PIN, HIGH);

  // Setup LPD8806 clock and data pins as output
  pinMode(LED_CLOCK_PIN, OUTPUT);
  digitalWrite(LED_CLOCK_PIN, LOW);
  pinMode(LED_DATA_PIN, OUTPUT);
  digitalWrite(LED_DATA_PIN, LOW);

  // Establish masks for quick access to LED strip
  clockPort = portOutputRegister(digitalPinToPort(LED_CLOCK_PIN));
  clockPinMask = digitalPinToBitMask(LED_CLOCK_PIN);
  dataPort = portOutputRegister(digitalPinToPort(LED_DATA_PIN));
  dataPinMask = digitalPinToBitMask(LED_DATA_PIN);

  // Set all LEDs in strip to black color
  clearLEDs();

  secondsToggle = false;
  hourMinuteToggle = false;

  // Startup Wire library
  Wire.begin();

  // Check for RTC operation
  if (! rtcRunning()) {
    if (DEBUG) {
      Serial.println("RTC is NOT running!");
    }
  }

  // Put RTC into 12 hour format
  set12HourFormat();

  // Set initial state of clock
  clockState = DISPATCH;

  // Read the buttons to allow hardware to settle
  while (readButtonState() != NEITHER) {
    ;
  }
}

/************************************************************************/
/********************** Loop during clock operation *********************/
/************************************************************************/

void loop() {

  BUTTON_STATE state;
  boolean changeMade;

  // Get the current time / date from the RTC
  getTime();
  getDate();

  if (DEBUG) {
    printDateTime();
  }

  // Clock runs as state machine
  switch (clockState) {

  case DISPATCH:
    {
      // Dispatch events in priority order

      // Clock setting mode ?
      if (readButtonState() == BOTH) {
        clockState = SETTING;
      } 
      // Zero and 30 minute event ? 
      else if ((minutes != latchedMinute) && ((minutes % 30) == 0)) {
        clockState = EVENT30;
      }
      // 15 minute event ? 
      else if ((minutes != latchedMinute) && (minutes != 0) && ((minutes % 15) == 0)) {
        clockState = EVENT15;
      }
      // 10 minute event ? 
      else if ((minutes != latchedMinute) && (minutes != 0) && ((minutes % 10) == 0)) {
        clockState = EVENT10;
      }  
      else  {
        clockState = RUNNING;
      }
      break;
    }

  case EVENT10:
    {
      latchedMinute = minutes;

      // Every event do the rotator pattern
      rotatorPattern();

      // Set next state
      clockState = DISPATCH;
      break;
    }  

  case EVENT15:
    {
      latchedMinute = minutes;

      // Every event display the full date
      displayFullDate();

      clockState = DISPATCH;
      break;
    }

  case EVENT30:
    {
      latchedMinute = minutes;

      // Every event do the brightening pattern
      brighteningPattern();

      // Set next state
      clockState = DISPATCH;
      break;
    }  

  case SETTING:
    {
      // Indicate we are in hour setting mode
      displayFlash(hourColor);

      // Display current hour setting
      displayHour();

      changeMade = false;

      // Read the state of the buttons
      state = readButtonState();

      // Does user want to leave hour setting mode ?
      while (state != BOTH) {
        if (state == PLUS) {
          hours++;
          if (hours > 12) {
            hours = 1;
            pmAm = ! pmAm;
          }  
          changeMade = true;
        }
        else if (state == MINUS) {
          hours--;
          if (hours < 1) {
            hours = 12;
            pmAm = ! pmAm;
          }
          changeMade = true;
        }
        if (changeMade) {
          // Display current hour setting
          displayHour();
          changeMade = false;
        }
        // Read the state of the buttons
        state = readButtonState();
      }

      // Indicate we are in min setting mode
      displayFlash(minuteColor);

      // Display current setting
      displayMinute();

      changeMade = false;

      // Read the state of the buttons
      state = readButtonState();

      // Does user want to leave minute setting mode ?
      while (state != BOTH) {
        if (state == PLUS) {
          minutes++;
          minutes = (minutes > 59) ? 0 : minutes;
          changeMade = true;
        }  
        else if (state == MINUS) {
          minutes--;
          minutes = (minutes == 255) ? 59 : minutes;
          changeMade = true;
        }
        if (changeMade) {
          // Display current setting
          displayMinute();
          changeMade = false;
          seconds = 0;
        }
        // Read the state of the buttons
        state = readButtonState();
      }

      // Indicate we are in month setting mode
      displayFlash(monthColor);

      // Display current setting
      displayMonth();

      changeMade = false;

      // Read the state of the buttons
      state = readButtonState();

      // Does user want to leave month setting mode ?
      while (state != BOTH) {
        if (state == PLUS) {
          month++;
          month = (month > 12) ? 1 : month;
          changeMade = true;
        }  
        else if (state == MINUS) {
          month--;
          month = (month < 1) ? 12 : month;
          changeMade = true;
        }
        if (changeMade) {
          // Display current setting
          displayMonth();
          changeMade = false;
        }
        // Read the state of the buttons
        state = readButtonState();
      }

      // Indicate we are in date setting mode
      displayFlash(dateColor);

      // Display current setting
      displayDate();

      changeMade = false;

      // Read the state of the buttons
      state = readButtonState();

      // Does user want to leave date setting mode ?
      while (state != BOTH) {
        if (state == PLUS) {
          date++;
          date = (date > 31) ? 1 : date;
          changeMade = true;
        }  
        else if (state == MINUS) {
          date--;
          date = (date < 1) ? 31 : date;
          changeMade = true;
        }
        if (changeMade) {
          // Display current setting
          displayDate();
          changeMade = false;
        }
        // Read the state of the buttons
        state = readButtonState();
      }

      // Indicate we are in year setting mode
      displayFlash(yearColor);

      // Display current setting
      displayYear();

      changeMade = false;

      // Read the state of the buttons
      state = readButtonState();

      // Does user want to leave year setting mode ?
      while (state != BOTH) {
        if (state == PLUS) {
          year++;
          year = (year > 23) ? 0 : year;
          changeMade = true;
        }  
        else if (state == MINUS) {
          year--;
          year = (year < 0) ? 23 : year;
          changeMade = true;
        }
        if (changeMade) {
          // Display current setting
          displayYear();
          changeMade = false;
        }
        // Read the state of the buttons
        state = readButtonState();
      }

      if (DEBUG) {
        Serial.println("****************************************");
        Serial.print("Newly set time/date: ");
        printDateTime();
        Serial.println("****************************************");
      }

      clearLEDs();

      // Update the time / date in the RTC
      setTime();
      setDate();

      // Set next state
      clockState = DISPATCH;
      break;
    }  

  case RUNNING:
    {
      // Display mode of operation

      // Toggle seconds indication
      if (secondsToggle) {
        setLEDColor(0, RED);
        setLEDColor(1, BLACK);
        show();
        secondsToggle = false;
      }  
      else  {
        setLEDColor(0, BLACK);
        setLEDColor(1, RED);
        show();
        secondsToggle = true;
      }
      // Set background color of LEDs
      displayBackground();

      // Display the minute value as closest 5 minute LED
      uint8_t displayMinute = (uint8_t) minutes / 5;

      // Hour and Minute on same time ?
      // If so alternate hour and minute color on LED
      if (displayMinute == hours) {

        if (hourMinuteToggle) {
          hourMinuteToggle = false;
          setLEDColor(HOUR_LED_MAP[hours], hourColor);
        }  
        else  {
          hourMinuteToggle = true;
          setLEDColor(HOUR_LED_MAP[displayMinute], minuteColor);
        }
      }  
      else  {
        // Hour and minutes display different
        hourMinuteToggle = false;
        setLEDColor(HOUR_LED_MAP[hours], hourColor);
        setLEDColor(HOUR_LED_MAP[displayMinute], minuteColor);
      }
      show();
    }
    delay(1000);

    // Set next state
    clockState = DISPATCH;
    break; 
  } 
}































































